using Microsoft.Data.SqlClient;
using System.Data;
using WinFormsLocalDB.PashaCode;

//===========================================================
//        :
// https://www.interestprograms.ru
//     
//===========================================================

namespace SQLDataGridViewDataTable
{
    public partial class Form1 : Form
    {
        readonly Color defaultColor;
        readonly Color newRowColor = Color.Yellow;
        //DataTable tableDefault = new();

        public Form1()
        {
            InitializeComponent();

            Init();

            defaultColor = dgvTable.DefaultCellStyle.BackColor;
        }

        private void Init()
        {
            string queryString = "SELECT * FROM INFORMATION_SCHEMA.TABLES";
            DataTable dt = ConnectDB.SqlQuery(queryString);
            dgvDatabase.DataSource = dt;
        }

        //   
        private void DgvDatabase_RowEnter(object sender, DataGridViewCellEventArgs e)
        {
            string queryString = "SELECT * FROM " + dgvDatabase.Rows[e.RowIndex].Cells["TABLE_NAME"].Value;
            DataTable dataTable = ConnectDB.SqlQuery(queryString);
            dgvTable.DataSource = dataTable;
        }


        private void DgvTable_CellValueChanged(object sender, DataGridViewCellEventArgs e)
        {
            DataTable dtSource = (DataTable)dgvTable.DataSource;

            //       ,  .
            if (dtSource.Rows.Count - 1 >= e.RowIndex && dtSource.Rows[e.RowIndex].RowState == DataRowState.Unchanged)
            {
                Update(dtSource, e.RowIndex, e.ColumnIndex);
            }
        }

        private void DgvTable_DataError(object sender, DataGridViewDataErrorEventArgs e)
        {
            //   .
            string error = " -  " + (e.RowIndex + 1) + " : " + e.Exception.Message;
            string action = " - \n  - ";

            //   .
            labelErrorMsg.Text = error;
            dgvTable.Rows[e.RowIndex].ErrorText = error;

            if (MessageBox.Show(error + "\n" + action, "!", MessageBoxButtons.YesNo, MessageBoxIcon.Warning) == DialogResult.No)
            {
                e.Cancel = true;
            }

        }

        private void DgvTable_CellValidating(object sender, DataGridViewCellValidatingEventArgs e)
        {
            if (dgvTable.Rows[e.RowIndex].IsNewRow == true)
            {
                return;
            }

            DataTable dtSource = (DataTable)dgvTable.DataSource;

            //     .
            Type type = dtSource.Columns[e.ColumnIndex].DataType;

            string? error = e.FormattedValue.ToString() switch
            {
                { Length: 0 } when dtSource.Columns[e.ColumnIndex].AllowDBNull == false => "    !",
                { Length: > 0 } when TryConvert(e.FormattedValue, type) == false => "   " + type.Name + "!",
                _ => null
            };

            if (error != null)
            {
                string action = "\n - \n  - ";
                //      .
                string errormsg = "!  " + (e.RowIndex + 1) + ",  " + (e.ColumnIndex + 1) + ": " + error;
                dgvTable.Rows[e.RowIndex].ErrorText = errormsg;
                labelErrorMsg.Text = errormsg;

                if (MessageBox.Show(error + "\n" + action, " !", MessageBoxButtons.YesNo, MessageBoxIcon.Warning) == DialogResult.No)
                {
                    e.Cancel = true;
                }
                else
                {
                    //    .
                    dgvTable.Rows[e.RowIndex].ErrorText = "";
                    labelErrorMsg.Text = "";

                    //    ,    .
                    dgvTable.CancelEdit();
                }
            }
        }

        private void DgvTable_CellValidated(object sender, DataGridViewCellEventArgs e)
        {
            //    .
            dgvTable.Rows[e.RowIndex].Cells[e.ColumnIndex].ErrorText = "";
            dgvTable.Rows[e.RowIndex].ErrorText = "";
            labelErrorMsg.Text = "";
        }

        private void DgvTable_RowValidating(object sender, DataGridViewCellCancelEventArgs e)
        {
            dgvTable.Rows[e.RowIndex].Cells[e.ColumnIndex].ErrorText = "";
            dgvTable.Rows[e.RowIndex].ErrorText = "";
            labelErrorMsg.Text = "";
            dgvTable.Rows[e.RowIndex].DefaultCellStyle.BackColor = defaultColor;

        }



        private void DgvTable_RowValidated(object sender, DataGridViewCellEventArgs e)
        {
            DataTable dtSource = (DataTable)dgvTable.DataSource;

            foreach (DataRow row in dtSource.Rows)
            {
                if (row.RowState == DataRowState.Added)
                {
                    Insert(dtSource, row);
                }
            }
        }


        private void DgvTable_RowEnter(object sender, DataGridViewCellEventArgs e)
        {
            if (dgvTable.Rows[e.RowIndex].IsNewRow == true)
            {
                //        .
                dgvTable.Rows[e.RowIndex].DefaultCellStyle.BackColor = newRowColor;

                //      ,
                //        .
                dgvTable.Rows[e.RowIndex].Cells["Id"].Value = "-1";
            }
        }


        private void DgvTable_KeyDown(object sender, KeyEventArgs e)
        {
            if (e.KeyCode == Keys.Delete)
            {
                int numberrowsdelete = dgvTable.SelectedRows.Count;

                DataTable dt = (DataTable)dgvTable.DataSource;

                if (MessageBox.Show(
                    " ?\n - " + numberrowsdelete,
                    " !",
                    MessageBoxButtons.YesNo, MessageBoxIcon.Exclamation) == DialogResult.Yes)
                {
                    string? queryString = null;
                    foreach (DataGridViewRow row in dgvTable.SelectedRows)
                    {
                        queryString += "DELETE FROM " + dt.TableName + " WHERE Id=" + row.Cells["Id"].Value.ToString() + ";";
                    }
                    if (queryString != null)
                        ConnectDB.SqlQuery(queryString);
                }
            }
        }

        #region 

        void Update(DataTable dtSource, int rowindex, int colindex)
        {
            string tablename = dtSource.TableName;
            object id = dtSource.Rows[rowindex]["Id"];
            string columnname = dtSource.Columns[colindex].ColumnName;
            object value = dtSource.Rows[rowindex][colindex];

            List<SqlParameter> parameters = new()
            {
                new SqlParameter()
                {
                    Direction = ParameterDirection.Input,
                    ParameterName = "Id",
                    Value = id
                },
                new SqlParameter()
                {
                    Direction = ParameterDirection.Input,
                    ParameterName = "Value",
                    Value = value
                }
            };

            string queryString = "UPDATE " + tablename + " SET [" + columnname + "]=@Value WHERE Id=@Id";
            ConnectDB.SqlQuery(queryString, parameters);
        }

        void Insert(DataTable dtSource, DataRow row)
        {
            List<SqlParameter> parameters = new();
            List<string> listnames = new();
            List<string> listvalues = new();

            //     .
            int count = 0;
            foreach (DataColumn col in row.Table.Columns)
            {
                if (col.AutoIncrement == false)
                {
                    listnames.Add("[" + col.ColumnName + "]");

                    SqlParameter parUri = new()
                    {
                        //     ,
                        //    .
                        //DbType = ConvertToDBType.Convert(dataTable.Columns[colindex].DataType),
                        Direction = ParameterDirection.Input,
                        ParameterName = "Param" + count,
                        Value = row[col.ColumnName]
                    };
                    parameters.Add(parUri);
                    count++;

                    listvalues.Add("@" + parUri.ParameterName);
                }
            }

            string names = string.Join(',', listnames);
            string values = string.Join(',', listvalues);

            //       .
            row.AcceptChanges();


            //   ,         Id.
            //    datagridview        .
            // https://docs.microsoft.com/ru-ru/sql/t-sql/functions/ident-current-transact-sql?view=sql-server-ver15
            string querystring = "INSERT INTO " + row.Table.TableName + " (" + names + ") VALUES(" + values + ");";
            querystring += "SELECT IDENT_CURRENT('" + row.Table.TableName + "');";
            DataTable lastId = ConnectDB.SqlQuery(querystring, parameters);

            //         .
            //       
            //     .

            //  
            // https://docs.microsoft.com/ru-ru/dotnet/csharp/nullable-references

            //    ,
            //    .
            dtSource.Columns["Id"]!.ReadOnly = false;
            row["Id"] = lastId.Rows[0][0];
            dtSource.Columns["Id"]!.ReadOnly = true;
        }

        public static bool TryConvert(object unknown, Type type)
        {
            try
            {
                object d = Convert.ChangeType(unknown, type);
            }
            catch (System.FormatException)
            {
                return false;
            }

            return true;
        }

        public void GetDefaultValues(DataTable dtSource)
        {
            //           
            string sq = "SELECT COLUMN_NAME, COLUMN_DEFAULT FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_NAME='" + dtSource.TableName + "' AND COLUMNPROPERTY(object_id('" + dtSource.TableName + "'), COLUMN_NAME, 'IsIdentity')=0;";
            DataTable dtSchema = ConnectDB.SqlQuery(sq);

            int count = 0;
            List<SqlParameter> parameters = new();
            List<string> listnames = new();
            List<string> listvalues = new();

            foreach (DataRow r in dtSchema.Rows)
            {
                if (r["COLUMN_DEFAULT"] == DBNull.Value)
                {
                    listnames.Add("[" + r["COLUMN_NAME"] + "]");

                    SqlParameter parUri = new()
                    {
                        Direction = ParameterDirection.Input,
                        ParameterName = "Param" + count,
                        Value = ""
                    };
                    parameters.Add(parUri);

                    listvalues.Add("@" + parUri.ParameterName);

                    count++;
                }
            }

            string names = string.Join(',', listnames);
            string values = string.Join(',', listvalues);


            // =====
            //  1
            //  
            //    
            //    
            //  .
            string sqlquery = "BEGIN TRANSACTION;";
            sqlquery += "INSERT INTO " + dtSource.TableName + " (" + names + ") VALUES(" + values + ");";
            string lastid = "SELECT IDENT_CURRENT('" + dtSource.TableName + "')";
            sqlquery += "SELECT * FROM " + dtSource.TableName + " WHERE id=(" + lastid + ");";
            sqlquery += "ROLLBACK;";
            DataTable tableDefault = ConnectDB.SqlQuery(sqlquery, parameters);

            //         .
            foreach (DataRow r in dtSchema.Rows)
            {
                foreach (DataColumn col in dtSource.Columns)
                {
                    if ((string)r["COLUMN_NAME"] == col.ColumnName &&
                        r["COLUMN_DEFAULT"] != DBNull.Value)
                    {
                        object o = tableDefault.Rows[0][col.ColumnName];
                        col.DefaultValue = o;
                    }
                }
            }


            //  2
            // 1.       .
            // 2.      .
            // 3.       .
            // 4.       .

        }


        #endregion

       
        private void dgvTable_DataSourceChanged(object sender, EventArgs e)
        {
            DataTable dtSource = (DataTable)dgvTable.DataSource;
            GetDefaultValues(dtSource);
        }

    }
}